@@ -29,7 +29,7 @@
 
 #include "precompiled.hpp"
 #include "select.hpp"
-#if defined ZMQ_USE_SELECT
+#if defined ZMQ_IOTHREAD_POLLER_USE_SELECT
 
 #if defined ZMQ_HAVE_WINDOWS
 #elif defined ZMQ_HAVE_HPUX
@@ -39,6 +39,10 @@
 #elif defined ZMQ_HAVE_OPENVMS
 #include <sys/types.h>
 #include <sys/time.h>
+#elif defined ZMQ_HAVE_VXWORKS
+#include <sys/types.h>
+#include <sys/time.h>
+#include <strings.h>
 #else
 #include <sys/select.h>
 #endif
@@ -47,29 +51,35 @@
 #include "config.hpp"
 #include "i_poll_events.hpp"
 
-zmq::select_t::select_t (const zmq::ctx_t &ctx_) :
-    ctx (ctx_),
+#include <algorithm>
+#include <limits>
+#include <climits>
+
+zmq::select_t::select_t (const zmq::thread_ctx_t &ctx_) :
+    worker_poller_base_t (ctx_),
 #if defined ZMQ_HAVE_WINDOWS
     //  Fine as long as map is not cleared.
-    current_family_entry_it (family_entries.end ()),
+    _current_family_entry_it (_family_entries.end ())
 #else
-    maxfd (retired_fd),
+    _max_fd (retired_fd)
 #endif
-    stopping (false)
 {
 #if defined ZMQ_HAVE_WINDOWS
     for (size_t i = 0; i < fd_family_cache_size; ++i)
-        fd_family_cache [i] = std::make_pair ((fd_t)retired_fd, 0);
+        _fd_family_cache [i] = std::make_pair ((fd_t)retired_fd, 0);
 #endif
 }
 
 zmq::select_t::~select_t ()
 {
-    worker.stop ();
+    stop_worker ();
 }
 
 zmq::select_t::handle_t zmq::select_t::add_fd (fd_t fd_, i_poll_events *events_)
 {
+    check_thread ();
+    zmq_assert (fd_ != retired_fd);
+
     fd_entry_t fd_entry;
     fd_entry.fd = fd_;
     fd_entry.events = events_;
@@ -77,14 +87,16 @@ zmq::select_t::handle_t zmq::select_t::add_fd (fd_t fd_, i_poll_events *events_)
 #if defined ZMQ_HAVE_WINDOWS
     u_short family = get_fd_family (fd_);
     wsa_assert (family != AF_UNSPEC);
-    family_entry_t &family_entry = family_entries [family];
+    family_entry_t &family_entry = _family_entries[family];
+#else
+    family_entry_t &family_entry = _family_entry;
 #endif
     family_entry.fd_entries.push_back (fd_entry);
     FD_SET (fd_, &family_entry.fds_set.error);
 
 #if !defined ZMQ_HAVE_WINDOWS
-    if (fd_ > maxfd)
-        maxfd = fd_;
+    if (fd_ > _max_fd)
+        _max_fd = fd_;
 #endif
 
     adjust_load (1);
@@ -93,11 +105,11 @@ zmq::select_t::handle_t zmq::select_t::add_fd (fd_t fd_, i_poll_events *events_)
 }
 
 zmq::select_t::fd_entries_t::iterator
-zmq::select_t::find_fd_entry_by_handle (fd_entries_t &fd_entries,
+zmq::select_t::find_fd_entry_by_handle (fd_entries_t &fd_entries_,
                                         handle_t handle_)
 {
     fd_entries_t::iterator fd_entry_it;
-    for (fd_entry_it = fd_entries.begin (); fd_entry_it != fd_entries.end ();
+    for (fd_entry_it = fd_entries_.begin (); fd_entry_it != fd_entries_.end ();
          ++fd_entry_it)
         if (fd_entry_it->fd == handle_)
             break;
@@ -112,13 +124,14 @@ void zmq::select_t::trigger_events (const fd_entries_t &fd_entries_,
     //  Size is cached to avoid iteration through recently added descriptors.
     for (fd_entries_t::size_type i = 0, size = fd_entries_.size ();
          i < size && event_count_ > 0; ++i) {
-        const fd_entry_t &current_fd_entry = fd_entries_ [i];
+        //  fd_entries_[i] may not be stored, since calls to
+        //  in_event/out_event may reallocate the vector
 
-        if (is_retired_fd (current_fd_entry))
+        if (is_retired_fd (fd_entries_[i]))
             continue;
 
-        if (FD_ISSET (current_fd_entry.fd, &local_fds_set_.read)) {
-            current_fd_entry.events->in_event ();
+        if (FD_ISSET (fd_entries_[i].fd, &local_fds_set_.read)) {
+            fd_entries_[i].events->in_event ();
             --event_count_;
         }
 
@@ -126,36 +139,41 @@ void zmq::select_t::trigger_events (const fd_entries_t &fd_entries_,
         //  was retired before, we would already have continued, and I
         //  don't see where it might have been modified
         //  And if rc == 0, we can break instead of continuing
-        if (is_retired_fd (current_fd_entry) || event_count_ == 0)
+        if (is_retired_fd (fd_entries_[i]) || event_count_ == 0)
             continue;
 
-        if (FD_ISSET (current_fd_entry.fd, &local_fds_set_.write)) {
-            current_fd_entry.events->out_event ();
+        if (FD_ISSET (fd_entries_[i].fd, &local_fds_set_.write)) {
+            fd_entries_[i].events->out_event ();
             --event_count_;
         }
 
         //  TODO: same as above
-        if (is_retired_fd (current_fd_entry) || event_count_ == 0)
+        if (is_retired_fd (fd_entries_[i]) || event_count_ == 0)
             continue;
 
-        if (FD_ISSET (current_fd_entry.fd, &local_fds_set_.error)) {
-            current_fd_entry.events->in_event ();
+        if (FD_ISSET (fd_entries_[i].fd, &local_fds_set_.error)) {
+            fd_entries_[i].events->in_event ();
             --event_count_;
         }
     }
 }
 
 #if defined ZMQ_HAVE_WINDOWS
-bool zmq::select_t::try_remove_fd_entry (
-  family_entries_t::iterator family_entry_it, zmq::fd_t &handle_)
+int zmq::select_t::try_retire_fd_entry (
+  family_entries_t::iterator family_entry_it_, zmq::fd_t &handle_)
 {
-    family_entry_t &family_entry = family_entry_it->second;
+    family_entry_t &family_entry = family_entry_it_->second;
 
     fd_entries_t::iterator fd_entry_it =
       find_fd_entry_by_handle (family_entry.fd_entries, handle_);
+
     if (fd_entry_it == family_entry.fd_entries.end ())
-        return false;
-    if (family_entry_it != current_family_entry_it) {
+        return 0;
+
+    fd_entry_t &fd_entry = *fd_entry_it;
+    zmq_assert (fd_entry.fd != retired_fd);
+
+    if (family_entry_it_ != _current_family_entry_it) {
         //  Family is not currently being iterated and can be safely
         //  modified in-place. So later it can be skipped without
         //  re-verifying its content.
@@ -163,105 +181,119 @@ bool zmq::select_t::try_remove_fd_entry (
     } else {
         //  Otherwise mark removed entries as retired. It will be cleaned up
         //  at the end of the iteration. See zmq::select_t::loop
-        fd_entry_it->fd = retired_fd;
-        family_entry.retired = true;
+        fd_entry.fd = retired_fd;
+        family_entry.has_retired = true;
     }
     family_entry.fds_set.remove_fd (handle_);
-    return true;
+    return 1;
 }
 #endif
 
 void zmq::select_t::rm_fd (handle_t handle_)
 {
+    check_thread ();
+    int retired = 0;
 #if defined ZMQ_HAVE_WINDOWS
     u_short family = get_fd_family (handle_);
     if (family != AF_UNSPEC) {
         family_entries_t::iterator family_entry_it =
-          family_entries.find (family);
+          _family_entries.find (family);
 
-        int removed = try_remove_fd_entry (family_entry_it, handle_);
-        assert (removed);
+        retired += try_retire_fd_entry (family_entry_it, handle_);
     } else {
         //  get_fd_family may fail and return AF_UNSPEC if the socket was not
         //  successfully connected. In that case, we need to look for the
         //  socket in all family_entries.
-        family_entries_t::iterator end = family_entries.end ();
+        family_entries_t::iterator end = _family_entries.end ();
         for (family_entries_t::iterator family_entry_it =
-               family_entries.begin ();
+               _family_entries.begin ();
              family_entry_it != end; ++family_entry_it) {
-            if (try_remove_fd_entry (family_entry_it, handle_))
+            if (retired += try_retire_fd_entry (family_entry_it, handle_)) {
                 break;
+            }
         }
     }
 #else
     fd_entries_t::iterator fd_entry_it =
-      find_fd_entry_by_handle (family_entry.fd_entries, handle_);
-    assert (fd_entry_it != fd_entries.end ());
+      find_fd_entry_by_handle (_family_entry.fd_entries, handle_);
+    assert (fd_entry_it != _family_entry.fd_entries.end ());
 
+    zmq_assert (fd_entry_it->fd != retired_fd);
     fd_entry_it->fd = retired_fd;
-    family_entry.fds_set.remove_fd (handle_);
+    _family_entry.fds_set.remove_fd (handle_);
 
-    if (handle_ == maxfd) {
-        maxfd = retired_fd;
-        for (fd_entry_it = family_entry.fd_entries.begin ();
-             fd_entry_it != family_entry.fd_entries.end (); ++fd_entry_it)
-            if (fd_entry_it->fd > maxfd)
-                maxfd = fd_entry_it->fd;
+    ++retired;
+
+    if (handle_ == _max_fd) {
+        _max_fd = retired_fd;
+        for (fd_entry_it = _family_entry.fd_entries.begin ();
+             fd_entry_it != _family_entry.fd_entries.end (); ++fd_entry_it)
+            if (fd_entry_it->fd > _max_fd)
+                _max_fd = fd_entry_it->fd;
     }
 
-    family_entry.retired = true;
+    _family_entry.has_retired = true;
 #endif
+    zmq_assert (retired == 1);
     adjust_load (-1);
 }
 
 void zmq::select_t::set_pollin (handle_t handle_)
 {
+    check_thread ();
 #if defined ZMQ_HAVE_WINDOWS
     u_short family = get_fd_family (handle_);
     wsa_assert (family != AF_UNSPEC);
-    family_entry_t &family_entry = family_entries [family];
+    family_entry_t &family_entry = _family_entries[family];
+#else
+    family_entry_t &family_entry = _family_entry;
 #endif
     FD_SET (handle_, &family_entry.fds_set.read);
 }
 
 void zmq::select_t::reset_pollin (handle_t handle_)
 {
+    check_thread ();
 #if defined ZMQ_HAVE_WINDOWS
     u_short family = get_fd_family (handle_);
     wsa_assert (family != AF_UNSPEC);
-    family_entry_t &family_entry = family_entries [family];
+    family_entry_t &family_entry = _family_entries[family];
+#else
+    family_entry_t &family_entry = _family_entry;
 #endif
     FD_CLR (handle_, &family_entry.fds_set.read);
 }
 
 void zmq::select_t::set_pollout (handle_t handle_)
 {
+    check_thread ();
 #if defined ZMQ_HAVE_WINDOWS
     u_short family = get_fd_family (handle_);
     wsa_assert (family != AF_UNSPEC);
-    family_entry_t &family_entry = family_entries [family];
+    family_entry_t &family_entry = _family_entries[family];
+#else
+    family_entry_t &family_entry = _family_entry;
 #endif
     FD_SET (handle_, &family_entry.fds_set.write);
 }
 
 void zmq::select_t::reset_pollout (handle_t handle_)
 {
+    check_thread ();
 #if defined ZMQ_HAVE_WINDOWS
     u_short family = get_fd_family (handle_);
     wsa_assert (family != AF_UNSPEC);
-    family_entry_t &family_entry = family_entries [family];
+    family_entry_t &family_entry = _family_entries[family];
+#else
+    family_entry_t &family_entry = _family_entry;
 #endif
     FD_CLR (handle_, &family_entry.fds_set.write);
 }
 
-void zmq::select_t::start ()
-{
-    ctx.start_thread (worker, worker_routine, this);
-}
-
 void zmq::select_t::stop ()
 {
-    stopping = true;
+    check_thread ();
+    //  no-op... thread is stopped when no more fds or timers are registered
 }
 
 int zmq::select_t::max_fds ()
@@ -271,19 +303,33 @@ int zmq::select_t::max_fds ()
 
 void zmq::select_t::loop ()
 {
-    while (!stopping) {
+    while (true) {
         //  Execute any due timers.
-        int timeout = (int) execute_timers ();
+        int timeout = static_cast<int> (execute_timers ());
+
+        cleanup_retired ();
+
+#ifdef _WIN32
+        if (_family_entries.empty ()) {
+#else
+        if (_family_entry.fd_entries.empty ()) {
+#endif
+            zmq_assert (get_load () == 0);
+
+            if (timeout == 0)
+                break;
+
+            // TODO sleep for timeout
+            continue;
+        }
 
 #if defined ZMQ_HAVE_OSX
         struct timeval tv = {(long) (timeout / 1000), timeout % 1000 * 1000};
 #else
-        struct timeval tv = {(long) (timeout / 1000),
-                             (long) (timeout % 1000 * 1000)};
+        struct timeval tv = {static_cast<long> (timeout / 1000),
+                             static_cast<long> (timeout % 1000 * 1000)};
 #endif
 
-        int rc = 0;
-
 #if defined ZMQ_HAVE_WINDOWS
         /*
             On Windows select does not allow to mix descriptors from different
@@ -301,7 +347,8 @@ void zmq::select_t::loop ()
         */
 
         //  If there is just one family, there is no reason to use WSA events.
-        const bool use_wsa_events = family_entries.size () > 1;
+        int rc = 0;
+        const bool use_wsa_events = _family_entries.size () > 1;
         if (use_wsa_events) {
             // TODO: I don't really understand why we are doing this. If any of
             // the events was signaled, we will call select for each fd_family
@@ -313,8 +360,8 @@ void zmq::select_t::loop ()
             wsa_events_t wsa_events;
 
             for (family_entries_t::iterator family_entry_it =
-                   family_entries.begin ();
-                 family_entry_it != family_entries.end (); ++family_entry_it) {
+                   _family_entries.begin ();
+                 family_entry_it != _family_entries.end (); ++family_entry_it) {
                 family_entry_t &family_entry = family_entry_it->second;
 
                 for (fd_entries_t::iterator fd_entry_it =
@@ -327,18 +374,18 @@ void zmq::select_t::loop ()
                     if (FD_ISSET (fd, &family_entry.fds_set.read)
                         && FD_ISSET (fd, &family_entry.fds_set.write))
                         rc =
-                          WSAEventSelect (fd, wsa_events.events [3],
+                          WSAEventSelect (fd, wsa_events.events[3],
                                           FD_READ | FD_ACCEPT | FD_CLOSE
                                             | FD_WRITE | FD_CONNECT | FD_OOB);
                     else if (FD_ISSET (fd, &family_entry.fds_set.read))
-                        rc = WSAEventSelect (fd, wsa_events.events [0],
+                        rc = WSAEventSelect (fd, wsa_events.events[0],
                                              FD_READ | FD_ACCEPT | FD_CLOSE
                                                | FD_OOB);
                     else if (FD_ISSET (fd, &family_entry.fds_set.write))
-                        rc = WSAEventSelect (fd, wsa_events.events [1],
+                        rc = WSAEventSelect (fd, wsa_events.events[1],
                                              FD_WRITE | FD_CONNECT | FD_OOB);
                     else if (FD_ISSET (fd, &family_entry.fds_set.error))
-                        rc = WSAEventSelect (fd, wsa_events.events [2], FD_OOB);
+                        rc = WSAEventSelect (fd, wsa_events.events[2], FD_OOB);
                     else
                         rc = 0;
 
@@ -355,10 +402,10 @@ void zmq::select_t::loop ()
                 continue;
         }
 
-        for (current_family_entry_it = family_entries.begin ();
-             current_family_entry_it != family_entries.end ();
-             ++current_family_entry_it) {
-            family_entry_t &family_entry = current_family_entry_it->second;
+        for (_current_family_entry_it = _family_entries.begin ();
+             _current_family_entry_it != _family_entries.end ();
+             ++_current_family_entry_it) {
+            family_entry_t &family_entry = _current_family_entry_it->second;
 
 
             if (use_wsa_events) {
@@ -371,7 +418,7 @@ void zmq::select_t::loop ()
             }
         }
 #else
-        select_family_entry (family_entry, maxfd, timeout > 0, tv);
+        select_family_entry (_family_entry, _max_fd + 1, timeout > 0, tv);
 #endif
     }
 }
@@ -401,18 +448,7 @@ void zmq::select_t::select_family_entry (family_entry_t &family_entry_,
 
     trigger_events (fd_entries, local_fds_set, rc);
 
-    if (family_entry_.retired) {
-        family_entry_.retired = false;
-        family_entry_.fd_entries.erase (std::remove_if (fd_entries.begin (),
-                                                        fd_entries.end (),
-                                                        is_retired_fd),
-                                        family_entry_.fd_entries.end ());
-    }
-}
-
-void zmq::select_t::worker_routine (void *arg_)
-{
-    ((select_t *) arg_)->loop ();
+    cleanup_retired (family_entry_);
 }
 
 zmq::select_t::fds_set_t::fds_set_t ()
@@ -477,11 +513,43 @@ void zmq::select_t::fds_set_t::remove_fd (const fd_t &fd_)
     FD_CLR (fd_, &error);
 }
 
-bool zmq::select_t::is_retired_fd (const fd_entry_t &entry)
+bool zmq::select_t::cleanup_retired (family_entry_t &family_entry_)
 {
-    return (entry.fd == retired_fd);
+    if (family_entry_.has_retired) {
+        family_entry_.has_retired = false;
+        family_entry_.fd_entries.erase (
+          std::remove_if (family_entry_.fd_entries.begin (),
+                          family_entry_.fd_entries.end (), is_retired_fd),
+          family_entry_.fd_entries.end ());
+    }
+    return family_entry_.fd_entries.empty ();
 }
 
+void zmq::select_t::cleanup_retired ()
+{
+#ifdef _WIN32
+    for (family_entries_t::iterator it = _family_entries.begin ();
+         it != _family_entries.end ();) {
+        if (cleanup_retired (it->second))
+            it = _family_entries.erase (it);
+        else
+            ++it;
+    }
+#else
+    cleanup_retired (_family_entry);
+#endif
+}
+
+bool zmq::select_t::is_retired_fd (const fd_entry_t &entry_)
+{
+    return (entry_.fd == retired_fd);
+}
+
+zmq::select_t::family_entry_t::family_entry_t () : has_retired (false)
+{
+}
+
+
 #if defined ZMQ_HAVE_WINDOWS
 u_short zmq::select_t::get_fd_family (fd_t fd_)
 {
@@ -489,7 +557,7 @@ u_short zmq::select_t::get_fd_family (fd_t fd_)
     // for the same sockets, and determine_fd_family is expensive
     size_t i;
     for (i = 0; i < fd_family_cache_size; ++i) {
-        const std::pair<fd_t, u_short> &entry = fd_family_cache [i];
+        const std::pair<fd_t, u_short> &entry = _fd_family_cache[i];
         if (entry.first == fd_) {
             return entry.second;
         }
@@ -500,11 +568,11 @@ u_short zmq::select_t::get_fd_family (fd_t fd_)
     std::pair<fd_t, u_short> res =
       std::make_pair (fd_, determine_fd_family (fd_));
     if (i < fd_family_cache_size) {
-        fd_family_cache [i] = res;
+        _fd_family_cache[i] = res;
     } else {
         // just overwrite a random entry
         // could be optimized by some LRU strategy
-        fd_family_cache [rand () % fd_family_cache_size] = res;
+        _fd_family_cache[rand () % fd_family_cache_size] = res;
     }
 
     return res.second;
@@ -519,48 +587,43 @@ u_short zmq::select_t::determine_fd_family (fd_t fd_)
     int type;
     int type_length = sizeof (int);
 
-    int rc =
-      getsockopt (fd_, SOL_SOCKET, SO_TYPE, (char *) &type, &type_length);
+    int rc = getsockopt (fd_, SOL_SOCKET, SO_TYPE,
+                         reinterpret_cast<char *> (&type), &type_length);
 
     if (rc == 0) {
         if (type == SOCK_DGRAM)
             return AF_INET;
-        else {
-            rc = getsockname (fd_, (sockaddr *) &addr, &addr_size);
 
-            //  AF_INET and AF_INET6 can be mixed in select
-            //  TODO: If proven otherwise, should simply return addr.sa_family
-            if (rc != SOCKET_ERROR)
-                return addr.ss_family == AF_INET6 ? AF_INET : addr.ss_family;
-        }
+        rc =
+          getsockname (fd_, reinterpret_cast<sockaddr *> (&addr), &addr_size);
+
+        //  AF_INET and AF_INET6 can be mixed in select
+        //  TODO: If proven otherwise, should simply return addr.sa_family
+        if (rc != SOCKET_ERROR)
+            return addr.ss_family == AF_INET6 ? AF_INET : addr.ss_family;
     }
 
     return AF_UNSPEC;
 }
 
-zmq::select_t::family_entry_t::family_entry_t () : retired (false)
-{
-}
-
-
 zmq::select_t::wsa_events_t::wsa_events_t ()
 {
-    events [0] = WSACreateEvent ();
-    wsa_assert (events [0] != WSA_INVALID_EVENT);
-    events [1] = WSACreateEvent ();
-    wsa_assert (events [1] != WSA_INVALID_EVENT);
-    events [2] = WSACreateEvent ();
-    wsa_assert (events [2] != WSA_INVALID_EVENT);
-    events [3] = WSACreateEvent ();
-    wsa_assert (events [3] != WSA_INVALID_EVENT);
+    events[0] = WSACreateEvent ();
+    wsa_assert (events[0] != WSA_INVALID_EVENT);
+    events[1] = WSACreateEvent ();
+    wsa_assert (events[1] != WSA_INVALID_EVENT);
+    events[2] = WSACreateEvent ();
+    wsa_assert (events[2] != WSA_INVALID_EVENT);
+    events[3] = WSACreateEvent ();
+    wsa_assert (events[3] != WSA_INVALID_EVENT);
 }
 
 zmq::select_t::wsa_events_t::~wsa_events_t ()
 {
-    wsa_assert (WSACloseEvent (events [0]));
-    wsa_assert (WSACloseEvent (events [1]));
-    wsa_assert (WSACloseEvent (events [2]));
-    wsa_assert (WSACloseEvent (events [3]));
+    wsa_assert (WSACloseEvent (events[0]));
+    wsa_assert (WSACloseEvent (events[1]));
+    wsa_assert (WSACloseEvent (events[2]));
+    wsa_assert (WSACloseEvent (events[3]));
 }
 #endif
 
